"""Common CLI workflow functions"""


import click

from .balsamic.base import balsamic as balsamic_cmd
from .fluffy.base import fluffy
from .microsalt.base import microsalt
from .mip_dna.base import mip_dna
from .mip_rna.base import mip_rna
from .mutant.base import mutant


@click.group()
def workflow():
    """Workflows commands"""


workflow.add_command(balsamic_cmd)
workflow.add_command(microsalt)
workflow.add_command(mip_dna)
workflow.add_command(mip_rna)
workflow.add_command(fluffy)
workflow.add_command(mutant)



"""Priority specific constants"""
from cgmodels.cg.constants import StrEnum

PRIORITY_MAP = {"research": 0, "standard": 1, "priority": 2, "express": 3, "clinical trials": 4}

REV_PRIORITY_MAP = {value: key for key, value in PRIORITY_MAP.items()}

PRIORITY_OPTIONS = list(PRIORITY_MAP.keys())


class SlurmQos(StrEnum):
    LOW: str = "low"
    NORMAL: str = "normal"
    HIGH: str = "high"



"""CLI for deploying tools with CG"""

import logging
from pathlib import Path

import click
from cg.apps.shipping import ShippingAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.group()
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def deploy(context: CGConfig, dry_run: bool):
    """Deploy tools with CG. Use --help to see what tools are available"""
    LOG.info("Running CG DEPLOY")
    shipping_api = context.shipping_api
    shipping_api.set_dry_run(dry_run)


@click.command(name="fluffy")
@click.pass_obj
def deploy_fluffy_cmd(context: CGConfig):
    """Deploy the fluffy tool"""
    LOG.info("Deploying fluffy with CG")
    fluffy_config: Path = Path(context.fluffy.deploy_config)
    shipping_api: ShippingAPI = context.shipping_api
    shipping_api.deploy(app_name="fluffy", app_config=fluffy_config)


@click.command(name="genotype")
@click.pass_obj
def deploy_genotype_cmd(context: CGConfig):
    """Deploy the genotype tool"""
    LOG.info("Deploying genotype with CG")
    shipping_api: ShippingAPI = context.shipping_api
    shipping_api.deploy(app_name="genotype")


@click.command(name="hermes")
@click.pass_obj
def deploy_hermes_cmd(context: CGConfig):
    """Deploy the hermes tool"""
    LOG.info("Deploying hermes with CG")
    hermes_config: Path = Path(context.hermes.deploy_config)
    shipping_api: ShippingAPI = context.shipping_api
    shipping_api.deploy(app_name="hermes", app_config=hermes_config)


@click.command(name="loqusdb")
@click.pass_obj
def deploy_loqusdb_cmd(context: CGConfig):
    """Deploy the LoqusDB tool"""
    LOG.info("Deploying LoqusDB with CG")
    shipping_api: ShippingAPI = context.shipping_api
    shipping_api.deploy(app_name="loqusdb")


@click.command(name="scout")
@click.pass_obj
def deploy_scout_cmd(context: CGConfig):
    """Deploy the scout tool"""
    LOG.info("Deploying scout-browser with CG")
    scout_config: Path = Path(context.scout.deploy_config)
    shipping_api: ShippingAPI = context.shipping_api
    shipping_api.deploy(app_name="scout", app_config=scout_config)


@click.command(name="shipping")
@click.pass_obj
def deploy_shipping_cmd(context: CGConfig):
    """Deploy the shipping tool"""
    LOG.info("Deploying shipping with CG")
    shipping_api: ShippingAPI = context.shipping_api
    shipping_api.deploy(app_name="shipping")


deploy.add_command(deploy_fluffy_cmd)
deploy.add_command(deploy_genotype_cmd)
deploy.add_command(deploy_hermes_cmd)
deploy.add_command(deploy_loqusdb_cmd)
deploy.add_command(deploy_scout_cmd)
deploy.add_command(deploy_shipping_cmd)



import logging
import click

from cg.meta.upload.nipt.models import StatinaUploadFiles
from cg.meta.upload.nipt.nipt import NiptUploadAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.group()
def statina():
    """Upload NIPT result files to Statina"""
    pass


@statina.command("case")
@click.argument("case_id", required=True)
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def batch(configs: CGConfig, case_id: str, dry_run: bool):
    """Loading batch into the NIPT database"""

    LOG.info("*** Statina UPLOAD START ***")

    nipt_upload_api = NiptUploadAPI(configs)
    nipt_upload_api.set_dry_run(dry_run=dry_run)
    statina_files: StatinaUploadFiles = nipt_upload_api.get_statina_files(case_id=case_id)
    if dry_run:
        LOG.info(f"Found file paths for statina upload: {statina_files.json(exclude_none=True)}")
    else:
        nipt_upload_api.upload_to_statina_database(statina_files=statina_files)



"""Parse the MIP qc_metric file"""


def set_bamstats_metrics(file_metrics: dict, sample_data: dict) -> None:
    """Set bamstats metrics. Sum the total mapped reads for all sample BAM files"""
    total_reads = sample_data.get("reads", 0)
    sample_data["reads"] = int(file_metrics["bamstats"]["raw_total_sequences"]) + total_reads

    total_mapped = sample_data.get("total_mapped", 0)
    sample_data["total_mapped"] = int(file_metrics["bamstats"]["reads_mapped"]) + total_mapped


def set_chanjo_sexcheck_metrics(file_metrics: dict, sample_data: dict) -> None:
    """Set chanjo_sexcheck metrics"""
    sample_data["predicted_sex"] = file_metrics["chanjo_sexcheck"]["gender"]


def set_collecthsmetrics_metrics(file_metrics: dict, sample_data: dict) -> None:
    """Set collecthsmetrics metrics"""
    hs_metrics = file_metrics["collecthsmetrics"]["header"]["data"]
    sample_data["at_dropout"] = float(hs_metrics["AT_DROPOUT"])
    sample_data["completeness_target"] = {
        10: float(hs_metrics["PCT_TARGET_BASES_10X"]),
        20: float(hs_metrics["PCT_TARGET_BASES_20X"]),
        50: float(hs_metrics["PCT_TARGET_BASES_50X"]),
        100: float(hs_metrics["PCT_TARGET_BASES_100X"]),
    }
    sample_data["gc_dropout"] = float(hs_metrics["GC_DROPOUT"])
    sample_data["target_coverage"] = float(hs_metrics["MEAN_TARGET_COVERAGE"])


def set_collectmultiplemetrics_metrics(file_metrics: dict, sample_data: dict) -> None:
    """Set collectmultiplemetrics metrics"""
    mm_metrics = file_metrics["collectmultiplemetrics"]["header"]["pair"]
    sample_data["strand_balance"] = float(mm_metrics["STRAND_BALANCE"])


def set_collectmultiplemetricsinsertsize_metrics(file_metrics: dict, sample_data: dict) -> None:
    """Set collectmultiplemetricsinsertsize metrics"""
    mm_insert_metrics = file_metrics["collectmultiplemetricsinsertsize"]["header"]["data"]
    sample_data["median_insert_size"] = int(mm_insert_metrics["MEDIAN_INSERT_SIZE"])
    sample_data["insert_size_standard_deviation"] = float(mm_insert_metrics["STANDARD_DEVIATION"])


def set_markduplicates_metrics(file_metrics: dict, sample_data: dict) -> None:
    """Set markduplicates metrics"""
    sample_data["duplicates"] = float(file_metrics["markduplicates"]["fraction_duplicates"])


def get_sample_metrics(sample_metrics: dict, sample_data: dict) -> dict:
    """Get tool qc metrics from sample metrics"""
    get_metrics = {
        "bamstats": set_bamstats_metrics,
        "chanjo_sexcheck": set_chanjo_sexcheck_metrics,
        "collecthsmetrics": set_collecthsmetrics_metrics,
        "collectmultiplemetrics": set_collectmultiplemetrics_metrics,
        "collectmultiplemetricsinsertsize": set_collectmultiplemetricsinsertsize_metrics,
        "markduplicates": set_markduplicates_metrics,
    }

    for file_metrics in sample_metrics.values():

        for tool in file_metrics:

            if get_metrics.get(tool):
                get_metrics[tool](file_metrics=file_metrics, sample_data=sample_data)
    return sample_data


def parse_qcmetrics(metrics: dict) -> dict:
    """Parse MIP qc metrics file
    Args:
        metrics (dict): raw YAML input from MIP qc metrics file
    Returns:
        dict: parsed qc metrics metrics
    """
    qc_metric = {}

    for sample_id, sample_metrics in metrics["sample"].items():

        sample_data = {
            "id": sample_id,
        }
        sample_data = get_sample_metrics(sample_metrics=sample_metrics, sample_data=sample_data)
        sample_data["mapped"] = sample_data["total_mapped"] / sample_data["reads"]
        qc_metric[sample_id] = sample_data
    return qc_metric



""" Test the status database report helper"""
from datetime import datetime, timedelta

from cg.meta.report.report_helper import ReportHelper


def test_get_previous_report_version_when_only_one(store, helpers):

    # GIVEN two analyses for the given case
    first_analysis = helpers.add_analysis(store)

    # WHEN fetching previous_report_version
    report_version = ReportHelper.get_previous_report_version(first_analysis)

    # THEN the version should be None
    assert not report_version


def test_get_previous_report_version_when_two(store, helpers):

    # GIVEN two analyses for the given case
    yesterday = datetime.now() - timedelta(days=1)
    first_analysis = helpers.add_analysis(store, completed_at=datetime.now())
    helpers.add_analysis(store, first_analysis.family, completed_at=yesterday)

    # WHEN fetching previous_report_version
    report_version = ReportHelper.get_previous_report_version(first_analysis)

    # THEN the version should be 1
    assert report_version == 1


def test_first_analysis_when_only_one(store, helpers):

    # GIVEN one analysis for the given case
    analysis = helpers.add_analysis(store)
    assert len(analysis.family.analyses) == 1

    # WHEN fetching report_version
    report_version = ReportHelper.get_report_version(analysis)

    # THEN the version should be 1
    assert report_version == 1


def test_first_analysis_when_two(store, helpers):

    # GIVEN two analyses for the given case
    yesterday = datetime.now() - timedelta(days=1)
    first_analysis = helpers.add_analysis(store, completed_at=datetime.now())
    second_analysis = helpers.add_analysis(store, first_analysis.family, completed_at=yesterday)

    # WHEN fetching report_version
    report_version = ReportHelper.get_report_version(second_analysis)

    # THEN the version should be 1
    assert report_version == 1


def test_second_analysis_when_two(store, helpers):

    # GIVEN two analyses for the given case
    yesterday = datetime.now() - timedelta(days=1)
    first_analysis = helpers.add_analysis(store, completed_at=datetime.now())
    second_analysis = helpers.add_analysis(store, first_analysis.family, completed_at=yesterday)
    assert first_analysis.family.analyses.index(second_analysis) == 1
    assert first_analysis.family.analyses.index(first_analysis) == 0

    # WHEN fetching report_version
    report_version = ReportHelper.get_report_version(second_analysis)

    # THEN the version should be 1
    assert report_version == 1



import logging
from typing import Dict, Iterable, Optional

import sqlalchemy
from cg.apps.cgstats.crud import find
from cg.apps.cgstats.db import models as stats_models
from cg.apps.cgstats.demux_sample import DemuxSample, get_demux_samples
from cg.apps.cgstats.stats import StatsAPI
from cg.models.demultiplex.demux_results import DemuxResults, LogfileParameters
from cgmodels.demultiplex.sample_sheet import NovaSeqSample, SampleSheet

LOG = logging.getLogger(__name__)


def create_support_parameters(
    manager: StatsAPI, demux_results: DemuxResults
) -> stats_models.Supportparams:
    logfile_parameters: LogfileParameters = demux_results.get_logfile_parameters()
    support_parameters = manager.Supportparams()
    support_parameters.document_path = str(
        demux_results.results_dir
    )  # This is the unaligned directory
    support_parameters.idstring = logfile_parameters.id_string
    support_parameters.program = logfile_parameters.program
    support_parameters.commandline = logfile_parameters.command_line
    support_parameters.sampleconfig_path = str(demux_results.sample_sheet_path)
    support_parameters.sampleconfig = demux_results.sample_sheet_path.read_text()
    support_parameters.time = logfile_parameters.time
    manager.add(support_parameters)
    manager.flush()
    LOG.info("Creating new support parameters object %s", support_parameters)
    return support_parameters


def create_datasource(
    manager: StatsAPI, demux_results: DemuxResults, support_parameters_id: int
) -> stats_models.Datasource:
    datasource = manager.Datasource()
    datasource.runname = demux_results.run_name
    datasource.rundate = demux_results.run_date
    datasource.machine = demux_results.machine_name
    datasource.server = demux_results.demux_host
    datasource.document_path = str(demux_results.conversion_stats_path)
    datasource.document_type = "xml"
    datasource.time = sqlalchemy.func.now()
    datasource.supportparams_id = support_parameters_id

    manager.add(datasource)
    manager.flush()
    LOG.info("Creating new datasource object %s", datasource)
    return datasource


def create_flowcell(manager: StatsAPI, demux_results: DemuxResults) -> stats_models.Flowcell:
    flowcell = manager.Flowcell()
    flowcell.flowcellname = demux_results.flowcell.flowcell_id
    flowcell.flowcell_pos = demux_results.flowcell.flowcell_position
    flowcell.hiseqtype = "novaseq"
    flowcell.time = sqlalchemy.func.now()

    manager.add(flowcell)
    manager.flush()
    LOG.info("Creating new flowcell object %s", flowcell)
    return flowcell


def create_demux(manager: StatsAPI, flowcell_id: int, datasource_id: int) -> stats_models.Demux:
    demux: stats_models.Demux = manager.Demux()
    demux.flowcell_id = flowcell_id
    demux.datasource_id = datasource_id
    demux.basemask = ""
    demux.time = sqlalchemy.func.now()

    manager.add(demux)
    manager.flush()
    LOG.info("Creating new demux object %s", demux)
    return demux


def create_project(manager: StatsAPI, project_name: str) -> stats_models.Project:
    project: stats_models.Project = manager.Project()
    project.projectname = project_name
    project.time = sqlalchemy.func.now()
    manager.add(project)
    manager.flush()
    LOG.info("Creating new project object %s", project)
    return project


def create_sample(
    manager: StatsAPI, sample_id: str, barcode: str, project_id: int
) -> stats_models.Sample:
    sample: stats_models.Sample = manager.Sample()
    sample.project_id = project_id
    sample.samplename = sample_id
    sample.limsid = sample_id.split("_")[0]
    sample.barcode = barcode
    sample.time = sqlalchemy.func.now()

    manager.add(sample)
    manager.flush()
    return sample


def create_unaligned(
    manager: StatsAPI, demux_sample: DemuxSample, sample_id: int, demux_id: int
) -> stats_models.Unaligned:
    unaligned: stats_models.Unaligned = manager.Unaligned()
    unaligned.sample_id = sample_id
    unaligned.demux_id = demux_id
    unaligned.lane = demux_sample.lane
    unaligned.yield_mb = round(int(demux_sample.pass_filter_yield) / 1000000, 2)
    unaligned.passed_filter_pct = demux_sample.pass_filter_yield_pc
    unaligned.readcounts = demux_sample.pass_filter_clusters * 2
    unaligned.raw_clusters_per_lane_pct = demux_sample.raw_clusters_pc
    unaligned.perfect_indexreads_pct = (
        round(demux_sample.perfect_barcodes / demux_sample.barcodes * 100, 5)
        if demux_sample.barcodes
        else 0
    )
    unaligned.q30_bases_pct = demux_sample.pass_filter_Q30
    unaligned.mean_quality_score = demux_sample.pass_filter_qscore
    unaligned.time = sqlalchemy.func.now()

    manager.add(unaligned)
    manager.flush()
    return unaligned


def create_projects(manager: StatsAPI, project_names: Iterable[str]) -> Dict[str, int]:
    project_name_to_id: Dict[str, int] = {}
    for project_name in project_names:
        project_id: Optional[int] = find.get_project_id(project_name=project_name)
        if not project_id:
            project_object: stats_models.Project = create_project(
                manager=manager, project_name=project_name
            )
            project_id: int = project_object.project_id
        else:
            LOG.info("Project %s already exists", project_name)
        project_name_to_id[project_name] = project_id
    return project_name_to_id


def create_samples(
    manager: StatsAPI,
    demux_results: DemuxResults,
    project_name_to_id: Dict[str, int],
    demux_id: int,
) -> None:
    LOG.info("Creating samples for flowcell %s", demux_results.flowcell.flowcell_full_name)
    sample_sheet: SampleSheet = demux_results.flowcell.get_sample_sheet()
    demux_samples: Dict[int, Dict[str, DemuxSample]] = get_demux_samples(
        conversion_stats=demux_results.conversion_stats,
        demux_stats_path=demux_results.demux_stats_path,
        sample_sheet=sample_sheet,
    )
    sample: NovaSeqSample
    for sample in sample_sheet.samples:
        barcode = (
            sample.index
            if not sample.second_index
            else "+".join([sample.index, sample.second_index])
        )
        lane: int = sample.lane
        sample_id: Optional[int] = find.get_sample_id(sample_id=sample.sample_id, barcode=barcode)
        if sample.project == "indexcheck":
            LOG.debug("Skip adding indexcheck sample to database")
            continue
        project_id: int = project_name_to_id[sample.project]
        if not sample_id:
            sample_object: stats_models.Sample = create_sample(
                manager=manager, sample_id=sample.sample_id, barcode=barcode, project_id=project_id
            )
            sample_id: int = sample_object.sample_id

        unaligned_id: Optional[int] = find.get_unaligned_id(
            sample_id=sample_id, demux_id=demux_id, lane=sample.lane
        )
        if not unaligned_id:
            demux_sample: DemuxSample = demux_samples[lane][sample.sample_id]
            create_unaligned(
                manager=manager, demux_sample=demux_sample, sample_id=sample_id, demux_id=demux_id
            )


def create_novaseq_flowcell(manager: StatsAPI, demux_results: DemuxResults):
    """Add a novaseq flowcell to CG stats"""
    LOG.info("Adding flowcell information to cgstats")
    support_parameters_id: Optional[int] = find.get_support_parameters_id(
        demux_results=demux_results
    )
    if not support_parameters_id:
        support_parameters: stats_models.Supportparams = create_support_parameters(
            manager=manager, demux_results=demux_results
        )
        support_parameters_id: int = support_parameters.supportparams_id
    else:
        LOG.info("Support parameters already exists")
    datasource_id: Optional[int] = find.get_datasource_id(demux_results=demux_results)
    if not datasource_id:
        datasource_object: stats_models.Datasource = create_datasource(
            manager=manager,
            demux_results=demux_results,
            support_parameters_id=support_parameters_id,
        )
        datasource_id: int = datasource_object.datasource_id
    else:
        LOG.info("Data source already exists")
    flowcell_id: Optional[int] = find.get_flowcell_id(
        flowcell_name=demux_results.flowcell.flowcell_id
    )
    if not flowcell_id:
        flowcell: stats_models.Flowcell = create_flowcell(
            manager=manager, demux_results=demux_results
        )
        flowcell_id: int = flowcell.flowcell_id
    else:
        LOG.info("Flowcell already exists")

    demux_id: Optional[int] = find.get_demux_id(flowcell_object_id=flowcell_id)
    if not demux_id:
        demux_object: stats_models.Demux = create_demux(
            manager=manager, flowcell_id=flowcell_id, datasource_id=datasource_id
        )
        demux_id: int = demux_object.demux_id
    else:
        LOG.info("Demux object already exists")

    project_name_to_id = create_projects(manager=manager, project_names=demux_results.projects)
    create_samples(
        manager=manager,
        demux_results=demux_results,
        project_name_to_id=project_name_to_id,
        demux_id=demux_id,
    )
    manager.commit()



import logging
from pathlib import Path
from typing import Dict, List, Optional

from cg.exc import OrderError, OrderFormError
from cg.models.orders.orderform_schema import OrderCase, Orderform, OrderPool
from cg.models.orders.sample_base import OrderSample

LOG = logging.getLogger(__name__)


class OrderformParser:
    """Class to parse orderforms"""

    def __init__(self):
        self.samples: List[OrderSample] = []
        self.project_type: Optional[str] = None
        self.delivery_type: Optional[str] = None
        self.customer_id: Optional[str] = None
        self.order_comment: Optional[str] = None
        self.order_name: Optional[str] = None

    def parse_orderform(self, orderform_file: Path) -> None:
        """Parse the orderform information"""
        raise NotImplementedError

    def group_cases(self) -> Dict[str, List[OrderSample]]:
        """Group samples in cases."""
        LOG.info("Group samples under respective case")
        cases = {}
        for sample in self.samples:
            case_id = sample.case_id
            if not case_id:
                continue
            if case_id not in cases:
                cases[case_id] = []
            cases[case_id].append(sample)
        if cases:
            LOG.info("Found cases %s", ", ".join(cases.keys()))
        else:
            LOG.info("Could not find any cases")
        return cases

    def get_pools(self) -> List[OrderPool]:
        """create pools from samples

        Check that all samples from one pool has the same application
        """
        pools: Dict[str, OrderPool] = {}
        sample: OrderSample
        for sample in self.samples:
            pool_name = sample.pool
            if not pool_name:
                continue
            application = sample.application
            if pool_name not in pools:
                pools[pool_name] = OrderPool(
                    name=pool_name,
                    data_analysis=sample.data_analysis,
                    data_delivery=sample.data_delivery,
                    application=application,
                    samples=[sample],
                )
                continue
            # each pool must only have one application type
            pool = pools[pool_name]
            if str(pool.application) != application:
                raise OrderError(
                    f"different application in pool: {pool_name} - {[pools[pool_name].application, application]}"
                )
            pool.samples.append(sample)

        return [pools[pool_name] for pool_name in pools]

    def expand_case(self, case_id: str, case_samples: List[OrderSample]) -> OrderCase:
        """Fill-in information about case."""

        priorities = {sample.priority for sample in case_samples if sample.priority}
        if len(priorities) != 1:
            raise OrderFormError(f"multiple values for 'Priority' for case: {case_id}")

        gene_panels = set()
        for sample in case_samples:
            if not sample.panels:
                continue
            gene_panels.update(set(sample.panels))

        return OrderCase(
            name=case_id,
            samples=case_samples,
            require_qcok=any(sample.require_qcok for sample in case_samples),
            priority=priorities.pop(),
            panels=list(gene_panels),
        )

    def generate_orderform(self) -> Orderform:
        """Generate an orderform"""
        cases_map: Dict[str, List[OrderSample]] = self.group_cases()
        case_objs: List[OrderCase] = []
        for case_id in cases_map:
            case_objs.append(self.expand_case(case_id=case_id, case_samples=cases_map[case_id]))
        return Orderform(
            comment=self.order_comment,
            samples=self.samples,
            cases=case_objs,
            name=self.order_name,
            customer=self.customer_id,
            delivery_type=self.delivery_type,
            project_type=self.project_type,
            pools=self.get_pools(),
        )

    def __repr__(self):
        return (
            f"OrderformParser(project_type={self.project_type},delivery_type={self.delivery_type},customer_id="
            f"{self.customer_id},order_name={self.order_name})"
        )



from cg.apps.coverage import ChanjoAPI
from cg.apps.crunchy import CrunchyAPI
from cg.apps.gt import GenotypeAPI
from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.madeline.api import MadelineAPI
from cg.apps.mutacc_auto import MutaccAutoAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.apps.vogue import VogueAPI
from cg.meta.compress import CompressAPI
from cg.meta.upload.vogue import UploadVogueAPI
from cg.meta.workflow.prepare_fastq import PrepareFastqAPI
from cg.models.cg_config import CGConfig
from cg.store import Store


class MetaAPI:

    """MetaAPI class initializing all App APIs used within CG in non-conflicting manner"""

    def __init__(self, config: CGConfig):
        self.config = config
        self.chanjo_api: ChanjoAPI = config.chanjo_api
        self.crunchy_api: CrunchyAPI = config.crunchy_api
        self.genotype_api: GenotypeAPI = config.genotype_api
        self.hermes_api: HermesApi = config.hermes_api
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.lims_api: LimsAPI = config.lims_api
        self.madeline_api: MadelineAPI = config.madeline_api
        self.mutacc_auto_api: MutaccAutoAPI = config.mutacc_auto_api
        self.prepare_fastq_api: PrepareFastqAPI = PrepareFastqAPI(
            store=config.status_db,
            compress_api=CompressAPI(hk_api=config.housekeeper_api, crunchy_api=config.crunchy_api),
        )
        self.status_db: Store = config.status_db
        self.scout_api: ScoutAPI = config.scout_api
        self.trailblazer_api: TrailblazerAPI = config.trailblazer_api
        self.upload_vogue_api: UploadVogueAPI = UploadVogueAPI(
            genotype_api=config.genotype_api,
            vogue_api=config.vogue_api,
            store=config.status_db,
        )
        self.vogue_api: VogueAPI = config.vogue_api



from pathlib import Path

import pytest
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.models.slurm.sbatch import Sbatch
from cg.utils import Process
from tests.apps.crunchy.conftest import fixture_sbatch_process
from tests.mocks.process_mock import ProcessMock


@pytest.fixture(name="sbatch_parameters")
def fixture_sbatch_parameters(email_adress: str, slurm_account: str) -> Sbatch:
    config = {
        "job_name": "test",
        "account": slurm_account,
        "number_tasks": 3,
        "memory": 10,
        "log_dir": "path/to/dir",
        "email": email_adress,
        "hours": 2,
        "commands": "genmod",
    }
    return Sbatch.parse_obj(config)


@pytest.fixture(name="sbatch_content")
def fixture_sbatch_content(sbatch_parameters: Sbatch) -> str:
    api = SlurmAPI()
    return api.generate_sbatch_content(sbatch_parameters=sbatch_parameters)


@pytest.fixture(name="slurm_api")
def fixture_slurm_api(sbatch_process: ProcessMock) -> SlurmAPI:
    """Return a slurm api with the process mocked"""
    api = SlurmAPI()
    api.process = sbatch_process
    return api


@pytest.fixture(name="sbatch_path")
def fixture_sbatch_path(project_dir: Path) -> Path:
    return project_dir / "sbatch_file.sh"



class AnylysisStatus:
    CANCELLED: str = "cancelled"
    COMPLETED: str = "completed"
    ERROR: str = "error"
    FAILED: str = "failed"
    PENDING: str = "pending"
    RUNNING: str = "running"
    TIMEOUT: str = "timeout"



"""Import all constants here for easy access"""

from .compression import (
    FASTQ_DELTA,
    FASTQ_FIRST_READ_SUFFIX,
    FASTQ_SECOND_READ_SUFFIX,
    SPRING_SUFFIX,
)
from .constants import (
    CAPTUREKIT_CANCER_OPTIONS,
    CAPTUREKIT_OPTIONS,
    CASE_ACTIONS,
    COLLABORATORS,
    COMBOS,
    CONTAINER_OPTIONS,
    DEFAULT_CAPTURE_KIT,
    FLOWCELL_STATUS,
    PREP_CATEGORIES,
    SEX_OPTIONS,
    STATUS_OPTIONS,
    DataDelivery,
    Pipeline,
)
from .gene_panel import MASTER_LIST
from .paths import TMP_DIR
from .priority import PRIORITY_MAP, PRIORITY_OPTIONS, REV_PRIORITY_MAP
from .process import EXIT_FAIL, EXIT_SUCCESS, RETURN_SUCCESS
from .sample_sources import ANALYSIS_SOURCES, METAGENOME_SOURCES
from .symbols import NO_PARENT, SINGLE_QUOTE, SPACE
from .tags import HK_FASTQ_TAGS



from pathlib import Path
from typing import Optional
from pydantic import BaseModel


class StatinaUploadFiles(BaseModel):
    result_file: str
    multiqc_report: Optional[str]
    segmental_calls: Optional[str]



# -*- coding: utf-8 -*-
import json
import logging
import tempfile
from functools import wraps
from pathlib import Path
from typing import List, Optional

from cg.constants import ANALYSIS_SOURCES, METAGENOME_SOURCES
from cg.exc import DuplicateRecordError, OrderError, OrderFormError
from cg.meta.orders import OrdersAPI, OrderType
from cg.models.orders.order import OrderIn
from cg.store import models
from flask import Blueprint, abort, current_app, g, jsonify, make_response, request
from google.auth import jwt
from pydantic import ValidationError
from requests.exceptions import HTTPError
from sqlalchemy.orm import Query
from werkzeug.utils import secure_filename

from ..apps.orderform.excel_orderform_parser import ExcelOrderformParser
from ..apps.orderform.json_orderform_parser import JsonOrderformParser
from ..models.orders.orderform_schema import Orderform
from .ext import db, lims, osticket

LOG = logging.getLogger(__name__)
BLUEPRINT = Blueprint("api", __name__, url_prefix="/api/v1")


def public(route_function):
    @wraps(route_function)
    def public_endpoint(*args, **kwargs):
        return route_function(*args, **kwargs)

    public_endpoint.is_public = True
    return public_endpoint


@BLUEPRINT.before_request
def before_request():
    """Authorize API routes with JSON Web Tokens."""
    if request.method == "OPTIONS":
        return make_response(jsonify(ok=True), 204)

    endpoint_func = current_app.view_functions[request.endpoint]
    if getattr(endpoint_func, "is_public", None):
        return

    auth_header = request.headers.get("Authorization")
    if not auth_header:
        return abort(make_response(jsonify(message="no JWT token found on request"), 403))

    jwt_token = auth_header.split("Bearer ")[-1]
    try:
        user_data = jwt.decode(jwt_token, certs=current_app.config["GOOGLE_OAUTH_CERTS"])
    except ValueError:
        return abort(make_response(jsonify(message="outdated login certificate"), 403))

    user_obj = db.user(user_data["email"])
    if user_obj is None or not user_obj.order_portal_login:
        message = f"{user_data['email']} doesn't have access"
        return abort(make_response(jsonify(message=message), 403))

    g.current_user = user_obj


@BLUEPRINT.route("/submit_order/<order_type>", methods=["POST"])
def submit_order(order_type):
    """Submit an order for samples."""
    api = OrdersAPI(lims=lims, status=db, osticket=osticket)
    post_data: OrderIn = OrderIn.parse_obj(request.get_json())
    LOG.info("processing '%s' order: %s", order_type, post_data)
    try:
        result = api.submit(
            project=OrderType(order_type),
            order_in=post_data,
            user_name=g.current_user.name,
            user_mail=g.current_user.email,
        )
    except (DuplicateRecordError, OrderError) as error:
        return abort(make_response(jsonify(message=error.message), 401))
    except HTTPError as error:
        return abort(make_response(jsonify(message=str(error)), 401))

    return jsonify(
        project=result["project"], records=[record.to_dict() for record in result["records"]]
    )


@BLUEPRINT.route("/cases")
def cases():
    """Fetch cases."""
    records = db.cases(days=31)
    count = len(records)

    return jsonify(cases=records, total=count)


@BLUEPRINT.route("/families")
def families():
    """Fetch families."""
    if request.args.get("status") == "analysis":
        records = db.cases_to_mip_analyze()
        count = len(records)
    else:
        customer_objs: Optional[List[models.Customer]] = (
            None if g.current_user.is_admin else g.current_user.customers
        )
        case_query = db.families(
            enquiry=request.args.get("enquiry"),
            customers=customer_objs,
            action=request.args.get("action"),
        )
        count = case_query.count()
        records = case_query.limit(30)

    cases_data: List[dict] = [case_obj.to_dict(links=True) for case_obj in records]
    return jsonify(families=cases_data, total=count)


@BLUEPRINT.route("/families_in_customer_group")
def families_in_customer_group():
    """Fetch families in customer_group."""
    customer_objs: Optional[models.Customer] = (
        None if g.current_user.is_admin else g.current_user.customers
    )
    families_q: Query = db.families_in_customer_group(
        enquiry=request.args.get("enquiry"), customers=customer_objs
    )
    count = families_q.count()
    records = families_q.limit(30)
    data = [case_obj.to_dict(links=True) for case_obj in records]
    return jsonify(families=data, total=count)


@BLUEPRINT.route("/families/<family_id>")
def family(family_id):
    """Fetch a family with links."""
    case_obj = db.family(family_id)
    if case_obj is None:
        return abort(404)
    if not g.current_user.is_admin and (case_obj.customer not in g.current_user.customers):
        return abort(401)

    data = case_obj.to_dict(links=True, analyses=True)
    return jsonify(**data)


@BLUEPRINT.route("/families_in_customer_group/<family_id>")
def family_in_customer_group(family_id):
    """Fetch a family with links."""
    case_obj = db.family(family_id)
    if case_obj is None:
        return abort(404)
    elif not g.current_user.is_admin and (
        case_obj.customer.customer_group
        not in set(customer.customer_group for customer in g.current_user.customers)
    ):
        return abort(401)

    data = case_obj.to_dict(links=True, analyses=True)
    return jsonify(**data)


@BLUEPRINT.route("/samples")
def samples():
    """Fetch samples."""
    if request.args.get("status") and not g.current_user.is_admin:
        return abort(401)
    if request.args.get("status") == "incoming":
        samples_q = db.samples_to_receive()
    elif request.args.get("status") == "labprep":
        samples_q = db.samples_to_prepare()
    elif request.args.get("status") == "sequencing":
        samples_q = db.samples_to_sequence()
    else:
        customer_objs: Optional[models.Customer] = (
            None if g.current_user.is_admin else g.current_user.customers
        )
        samples_q = db.samples(enquiry=request.args.get("enquiry"), customers=customer_objs)
    limit = int(request.args.get("limit", 50))
    data = [sample_obj.to_dict() for sample_obj in samples_q.limit(limit)]
    return jsonify(samples=data, total=samples_q.count())


@BLUEPRINT.route("/samples_in_customer_group")
def samples_in_customer_group():
    """Fetch samples in a customer group."""
    customer_objs: Optional[models.Customer] = (
        None if g.current_user.is_admin else g.current_user.customers
    )
    samples_q = db.samples_in_customer_group(
        enquiry=request.args.get("enquiry"), customers=customer_objs
    )
    limit = int(request.args.get("limit", 50))
    data = [sample_obj.to_dict() for sample_obj in samples_q.limit(limit)]
    return jsonify(samples=data, total=samples_q.count())


@BLUEPRINT.route("/samples/<sample_id>")
def sample(sample_id):
    """Fetch a single sample."""
    sample_obj = db.sample(sample_id)
    if sample_obj is None:
        return abort(404)
    elif not g.current_user.is_admin and (sample_obj.customer not in g.current_user.customers):
        return abort(401)
    data = sample_obj.to_dict(links=True, flowcells=True)
    return jsonify(**data)


@BLUEPRINT.route("/samples_in_customer_group/<sample_id>")
def sample_in_customer_group(sample_id):
    """Fetch a single sample."""
    sample_obj = db.sample(sample_id)
    if sample_obj is None:
        return abort(404)
    elif not g.current_user.is_admin and (
        sample_obj.customer.customer_group
        not in set(customer.customer_group for customer in g.current_user.customers)
    ):
        return abort(401)
    data = sample_obj.to_dict(links=True, flowcells=True)
    return jsonify(**data)


@BLUEPRINT.route("/pools")
def pools():
    """Fetch pools."""
    customer_objs: Optional[models.Customer] = (
        None if g.current_user.is_admin else g.current_user.customers
    )
    pools_q = db.pools(customers=customer_objs, enquiry=request.args.get("enquiry"))
    data = [pool_obj.to_dict() for pool_obj in pools_q.limit(30)]
    return jsonify(pools=data, total=pools_q.count())


@BLUEPRINT.route("/pools/<pool_id>")
def pool(pool_id):
    """Fetch a single pool."""
    record = db.pool(pool_id)
    if record is None:
        return abort(404)
    elif not g.current_user.is_admin and (record.customer not in g.current_user.customers):
        return abort(401)
    return jsonify(**record.to_dict())


@BLUEPRINT.route("/flowcells")
def flowcells():
    """Fetch flowcells."""
    query = db.flowcells(status=request.args.get("status"), enquiry=request.args.get("enquiry"))
    data = [record.to_dict() for record in query.limit(50)]
    return jsonify(flowcells=data, total=query.count())


@BLUEPRINT.route("/flowcells/<flowcell_id>")
def flowcell(flowcell_id):
    """Fetch a single flowcell."""
    record = db.flowcell(flowcell_id)
    if record is None:
        return abort(404)
    return jsonify(**record.to_dict(samples=True))


@BLUEPRINT.route("/analyses")
def analyses():
    """Fetch analyses."""
    if request.args.get("status") == "delivery":
        analyses_q = db.analyses_to_deliver()
    elif request.args.get("status") == "upload":
        analyses_q = db.analyses_to_upload()
    else:
        analyses_q = db.Analysis.query
    data = [analysis_obj.to_dict() for analysis_obj in analyses_q.limit(30)]
    return jsonify(analyses=data, total=analyses_q.count())


@BLUEPRINT.route("/options")
def options():
    """Fetch various options."""
    customer_objs: Optional[models.Customer] = (
        db.Customer.query.all() if g.current_user.is_admin else g.current_user.customers
    )

    apptag_groups = {"ext": []}
    for application_obj in db.applications(archived=False):

        if not application_obj.versions:
            LOG.debug("Skipping application %s that doesn't have a price", application)
            continue

        if application_obj.is_external:
            apptag_groups["ext"].append(application_obj.tag)
        else:
            if application_obj.prep_category not in apptag_groups:
                apptag_groups[application_obj.prep_category] = []
            apptag_groups[application_obj.prep_category].append(application_obj.tag)

    source_groups = {"metagenome": METAGENOME_SOURCES, "analysis": ANALYSIS_SOURCES}

    return jsonify(
        customers=[
            {"text": f"{customer.name} ({customer.internal_id})", "value": customer.internal_id}
            for customer in customer_objs
        ],
        applications=apptag_groups,
        panels=[panel.abbrev for panel in db.panels()],
        organisms=[
            {
                "name": organism.name,
                "reference_genome": organism.reference_genome,
                "internal_id": organism.internal_id,
                "verified": organism.verified,
            }
            for organism in db.organisms()
        ],
        sources=source_groups,
        beds=[bed.name for bed in db.beds(hide_archived=True)],
    )


@BLUEPRINT.route("/me")
def me():
    """Fetch information about current user."""
    if not g.current_user.is_admin and not g.current_user.customers:
        return abort(401)

    return jsonify(user=g.current_user.to_dict())


@BLUEPRINT.route("/applications")
@public
def applications():
    """Fetch application tags."""
    query = db.applications(archived=False)
    data = [record.to_dict() for record in query]
    return jsonify(applications=data)


@BLUEPRINT.route("/applications/<tag>")
@public
def application(tag):
    """Fetch an application tag."""
    record = db.application(tag)
    if record is None:
        return abort(make_response(jsonify(message="application not found"), 404))
    return jsonify(**record.to_dict())


@BLUEPRINT.route("/orderform", methods=["POST"])
def orderform():
    """Parse an orderform/JSON export."""
    input_file = request.files.get("file")
    filename = secure_filename(input_file.filename)

    try:
        if filename.lower().endswith(".xlsx"):
            temp_dir = Path(tempfile.gettempdir())
            saved_path = str(temp_dir / filename)
            input_file.save(saved_path)
            order_parser = ExcelOrderformParser()
            order_parser.parse_orderform(excel_path=saved_path)
        else:
            json_data = json.load(input_file.stream)
            order_parser = JsonOrderformParser()
            order_parser.parse_orderform(order_data=json_data)
    except (OrderFormError, ValidationError) as error:
        if hasattr(error, "message"):
            message = error.message
        else:
            message = str(error)
        return abort(make_response(jsonify(message=message), 400))
    parsed_order: Orderform = order_parser.generate_orderform()

    return jsonify(**parsed_order.dict())



"""Parse statistics from the demultiplexing stats file"""
import logging
from pathlib import Path
from typing import Dict, Optional, Set
from xml.etree.ElementTree import Element, iterparse

from pydantic import BaseModel

LOG = logging.getLogger(__name__)


class SampleBarcodeStats(BaseModel):
    barcode_count: int
    perfect_barcode_count: int
    one_mismatch_barcode_count: Optional[int]


class DemuxStats:
    def __init__(self, demux_stats_path: Path):
        self.demux_stats_path: Path = demux_stats_path
        self._current_project = ""
        self._current_sample = ""
        self._current_barcode = ""
        self._current_lane: int = 0
        self._current_barcode_count = 0
        self._current_perfect_barcode_count = 0
        self._current_mismatch_barcode_count = 0
        self._skip_entry: bool = False
        self.flowcell_id = ""
        self.projects: Set[str] = set()
        self.samples: Set[str] = set()
        self.barcodes: Set[str] = set()
        self.barcode_to_sample: Dict[str, str] = {}
        self.lanes: Set[int] = set()
        self.lanes_to_barcode = {}
        self.barcode_to_lanes = {}
        self.parse_file()

    def create_entry(self) -> None:
        """Create a entry of SampleBarcodeStats and add it in a structured way"""
        entry: SampleBarcodeStats = SampleBarcodeStats(
            barcode_count=self._current_barcode_count,
            perfect_barcode_count=self._current_perfect_barcode_count,
            one_mismatch_barcode_count=self._current_mismatch_barcode_count,
        )
        LOG.debug("Creating entry %s", entry)
        if self._current_lane not in self.lanes_to_barcode:
            self.lanes_to_barcode[self._current_lane] = {}
        self.lanes_to_barcode[self._current_lane][self._current_barcode] = entry
        if self._current_barcode not in self.barcode_to_lanes:
            self.barcode_to_lanes[self._current_barcode] = {}
        self.barcode_to_lanes[self._current_barcode][self._current_lane] = entry

    def parse_file(self) -> None:
        """Parse a XML file with demux statistics"""
        LOG.info("Parsing demux stats file %s", self.demux_stats_path)
        for (event, node) in iterparse(str(self.demux_stats_path), ["start", "end"]):
            if event == "start":
                self.evaluate_start_event(node)
                continue
            if not self._skip_entry:
                self.evaluate_end_event(node)
            node.clear()

    def evaluate_start_event(self, node: Element) -> None:
        """Check what type a start event is and take the relevant action depending on type"""
        if node.tag == "Flowcell":
            self.set_flowcell_id(node)
        elif node.tag == "Project":
            self.set_project(node)
        elif node.tag == "Sample":
            if node.attrib["name"].lower() in ["all", "undetermined"]:
                self._skip_entry = True
                LOG.debug("Skip sample %s" % node.attrib["name"])
            self._skip_entry = False
            self.set_current_sample(node)
        elif node.tag == "Barcode":
            # Skip the barcode="All"
            if node.attrib["name"].lower() == "all":
                self._skip_entry = True
                return
            self._skip_entry = False
            self.set_current_barcode(node)
        elif node.tag == "Lane":
            self.set_current_lane(node)

    def evaluate_end_event(self, node: Element) -> None:
        """Check what type a end event is and take the relevant action depending on type"""
        # These needs to operate on end tags to ensure value exists
        if node.tag == "BarcodeCount":
            self._current_barcode_count = int(node.text)
        elif node.tag == "PerfectBarcodeCount":
            self._current_perfect_barcode_count = int(node.text.strip())
        elif node.tag == "OneMismatchBarcodeCount":
            self._current_mismatch_barcode_count = int(node.text)
        elif node.tag == "Lane":
            self.create_entry()

    def set_current_lane(self, node: Element) -> None:
        lane: int = int(node.attrib["number"].strip())
        LOG.debug("Set current lane to %s", lane)
        self._current_lane = lane
        self.lanes.add(lane)

    def set_current_barcode(self, node: Element) -> None:
        barcode_id = node.attrib["name"].strip()
        LOG.debug("Set current barcode %s", barcode_id)
        self._current_barcode = barcode_id
        self.barcodes.add(barcode_id)
        self.barcode_to_sample[barcode_id] = self._current_sample

    def set_current_sample(self, node: Element) -> None:
        sample_id = node.attrib["name"].strip()
        LOG.debug("Set current sample to %s", sample_id)
        self._current_sample = sample_id
        self.samples.add(sample_id)

    def set_project(self, node: Element) -> None:
        project: str = node.attrib["name"].strip()
        LOG.debug("Set current project to %s", project)
        self._current_project = project
        self.projects.add(project)

    def set_flowcell_id(self, node: Element) -> None:
        flowcell = node.attrib["flowcell-id"].strip()
        LOG.debug("Set flowcell id to %s", flowcell)
        self.flowcell_id = flowcell

    def __repr__(self):
        return (
            f"DemuxStats(flowcell_id={self.flowcell_id},projects={self.projects},lanes={self.lanes}"
        )
